//------------------------------------------------------------------------------
//	$Id: geodetic.js,v 1.11 2022/01/28 20:16:13 wolf Exp $
//------------------------------------------------------------------------------
//	Erklaerung:	https://www.netzwolf.info/ol2/geodetic.html
//------------------------------------------------------------------------------
//	Fragen, Wuensche, Bedenken, Anregungen?
//	<openlayers(%40)netzwolf.info>
//------------------------------------------------------------------------------

/*jsl:option explicit*/

'use strict';

var Geodetic = {

	//----------------------------------------------------------------------
	//	Constants
	//----------------------------------------------------------------------

	TWO_PI:		2 * Math.PI,
	RHO:		180.0 / Math.PI,

	//----------------------------------------------------------------------
	//	Auxiliary functions
	//----------------------------------------------------------------------

	toRadian: function (degrees) {

		return degrees / this.RHO;
	},

	toDegrees: function (radian) {

		return radian * this.RHO;
	},

	invguder: function (phi) {

		return Math.log (Math.tan(phi) + 1/Math.cos(phi));
	},

	//----------------------------------------------------------------------
	//	Auxiliary functions
	//----------------------------------------------------------------------

	normAlpha: function (alpha) {

		return alpha - Math.floor(alpha/this.TWO_PI) * this.TWO_PI;
	},

	normLambda: function (lambda) {

		return lambda - Math.floor(lambda/this.TWO_PI+0.5) * this.TWO_PI;
	},

	//----------------------------------------------------------------------
	//	spherical forward
	//----------------------------------------------------------------------

	sphericalForward: function (param) {

		//--------------------------------------------------------------
		//	extract parameter
		//--------------------------------------------------------------

		var lambda1   = param.lambda1;
		var phi1      = param.phi1;
		var alpha1    = param.alpha1;
		var delta     = param.delta;

		//--------------------------------------------------------------
		//	speedup
		//--------------------------------------------------------------

		var sinPhi1   = Math.sin (phi1);
		var cosPhi1   = Math.cos (phi1);
		var cosAlpha1 = Math.cos (alpha1);
		var sinDelta  = Math.sin (delta);
		var cosDelta  = Math.cos (delta);

		//--------------------------------------------------------------
		//	spherical geometry
		//--------------------------------------------------------------

		var signum = Math.sign (Math.sin (alpha1));

		var phi2 = Math.asin (
			sinPhi1 * cosDelta + cosPhi1 * sinDelta * cosAlpha1);

		var lambda2 = this.normLambda (lambda1 + signum * Math.acos (
			(cosPhi1 * cosDelta - sinPhi1 * sinDelta * cosAlpha1) /
				Math.cos(phi2)));

		var alpha2 = this.normAlpha (-signum * Math.acos (
			(sinPhi1 * sinDelta - cosPhi1 * cosDelta * cosAlpha1) /
				Math.cos(phi2)));

		//--------------------------------------------------------------
		//	result
		//--------------------------------------------------------------

		return {
			phi1: phi1, lambda1: lambda1, alpha1: alpha1,
			phi2: phi2, lambda2: lambda2, alpha2: alpha2,
			delta: delta
		};
	},

	//----------------------------------------------------------------------
	//	spherical inverse
	//----------------------------------------------------------------------

	sphericalInverse: function (param) {

		//--------------------------------------------------------------
		//	speedup
		//--------------------------------------------------------------

		var sinPhi1 = Math.sin (param.phi1);
		var sinPhi2 = Math.sin (param.phi2);
		var cosPhi1 = Math.cos (param.phi1);
		var cosPhi2 = Math.cos (param.phi2);

		//--------------------------------------------------------------
		//	spherical geometry
		//--------------------------------------------------------------

		var dLambda = this.normLambda(param.lambda2 - param.lambda1);

		var cosDelta =
			sinPhi1 * sinPhi2 +
			cosPhi1 * cosPhi2 * Math.cos(dLambda);

		var delta = Math.acos (cosDelta);

		var sinDelta = Math.sin (delta);

		//--------------------------------------------------------------
		//	directions, different at start and destination
		//--------------------------------------------------------------

		var alpha1 = 0;
		var alpha2 = 0;

		if (cosPhi1 && sinDelta) {

			alpha1 = this.normAlpha (Math.sign(dLambda) * Math.acos (
				(sinPhi2 - sinPhi1 * cosDelta) /
				(          cosPhi1 * sinDelta)));
		}

		if (cosPhi2 && sinDelta) {

			alpha2 = this.normAlpha (Math.sign(dLambda) * -Math.acos (
				(sinPhi1 - sinPhi2 * cosDelta) /
				(          cosPhi2 * sinDelta)));
		}

		//--------------------------------------------------------------
		//	result
		//--------------------------------------------------------------

		return {
			phi1: param.phi1, lambda1: param.lambda1, alpha1: alpha1,
			phi2: param.phi2, lambda2: param.lambda2, alpha2: alpha2,
			delta: delta
		};
	},

	//----------------------------------------------------------------------
	//	loxodromic forward
	//----------------------------------------------------------------------

	//----------------------------------------------------------------------
	//	loxodromic inverse
	//----------------------------------------------------------------------

	loxodromicInverse: function (param) {

		var deltaLambda = this.normLambda(param.lambda2 - param.lambda1);

		//--------------------------------------------------------------
		//	convert geographical to mercator
		//--------------------------------------------------------------

		var PHI1 = this.invguder (param.phi1);
		var PHI2 = this.invguder (param.phi2);

		//--------------------------------------------------------------
		//	loxodromic direction
		//--------------------------------------------------------------

		var alpha = Math.atan2 (deltaLambda, PHI2 - PHI1);

		//--------------------------------------------------------------
		//	loxodromic distance
		//--------------------------------------------------------------
		//	Parallel circles get a special treatment,
		//	because the common expression has a pole there
		//	(causing a division by zero).
		//--------------------------------------------------------------

		var cosAlpha = Math.cos (alpha);

		var delta = Math.abs (cosAlpha) >= 0.001 ?
			(param.phi2 - param.phi1) / cosAlpha :
			Math.abs (deltaLambda) * Math.cos (param.phi1);

		return {
			phi1:		param.phi1,
			lambad1:	param.lambda1,
			alpha1:		alpha,
			phi2:		param.phi2,
			lambad2:	param.lambda2,
			alpha2:		alpha,
			delta:		delta
		};
	},

	CLASS_NAME: 'Geodetic'
};

var GeodeticAnalyzer = {

	EARTH_RADIUS:	6378137.0,

	MAX_SEGMENT_LENGTH:	10000,
	MAX_SEGMENTS_PER_LINE:	100,

	math: Geodetic,

	projection: new OpenLayers.Projection('EPSG:4326'),

	textOrthodrome: 'Orthodrome',

	textOrthodromeInfo: 'Line k\xfcrzester Distanz, aber variabler Richtung;\n' +
			'wird in Mercator-Projektion gebogen dargestellt.',

	textLoxodrome: 'Loxodrome',

	textLoxodromeInfo: 'Line mit konstantem Steuerkurs;\n' +
			'wird in Mercator-Projektion als gerade Linie dargestellt.',

	textBearing: 'Richtung',
	textBearingStart: 'Richtung zu Beginn der Strecke',
	textBearingEnd:   'Richtung zum Ende der Strecke',
	textBearingConst: 'Richtung \xfcber die ganze Strecke',

	textDistance: 'Distanz',
	textDistanceNotMin: 'nicht die k\xfcrzestm\xf6gliche Entfernung',

	analyze: function (lonlats, onSuccess, onFailure) {

		if (lonlats.length<2) {

			onFailure ('Mindestens zwei Punkte ...');
			return;
		}

		var tableElement = document.createElement('table');
		var theadElement = document.createElement('thead');
		var tr, th, ae;

		tr = document.createElement('tr');


		ae = document.createElement('a');
		ae.appendChild(document.createTextNode(this.textOrthodrome));
		ae.setAttribute('title', this.textOrthodromeInfo);
		ae.setAttribute('rel', 'external');
		ae.setAttribute('target', '_blank');
		ae.setAttribute('href', '//de.wikipedia.org/wiki/Orthodrome');

		th = document.createElement('th');
		th.setAttribute('colspan', 5);
		th.appendChild(ae);
		tr.appendChild(th);

		ae = document.createElement('a');
		ae.appendChild(document.createTextNode(this.textLoxodrome));
		ae.setAttribute('title', this.textLoxodromeInfo);
		ae.setAttribute('rel', 'external');
		ae.setAttribute('target', '_blank');
		ae.setAttribute('href', '//de.wikipedia.org/wiki/Loxodrome');

		th = document.createElement('th');
		th.setAttribute('colspan', 3);
		th.appendChild(ae);
		tr.appendChild(th);

		theadElement.appendChild(tr);

		tr = document.createElement('tr');

		th = document.createElement('th');
		th.appendChild(document.createTextNode(this.textBearing));
		th.setAttribute('colspan', 2);
		th.setAttribute('title', this.textBearingStart);
		th.style.cursor='help';
		tr.appendChild(th);

		th = document.createElement('th');
		th.appendChild(document.createTextNode(this.textDistance));
		tr.appendChild(th);

		th = document.createElement('th');
		th.appendChild(document.createTextNode(this.textBearing));
		th.setAttribute('colspan', 2);
		th.setAttribute('title', this.textBearingEnd);
		th.style.cursor='help';
		tr.appendChild(th);

		th = document.createElement('th');
		th.appendChild(document.createTextNode(this.textBearing));
		th.setAttribute('colspan', 2);
		th.setAttribute('title', this.textBearingConst);
		th.style.cursor='help';
		tr.appendChild(th);

		th = document.createElement('th');
		th.appendChild(document.createTextNode(this.textDistance));
		th.setAttribute('title', this.textDistanceNotMin);
		tr.appendChild(th);

		theadElement.appendChild(tr);

		tableElement.appendChild (theadElement);

		//--------------------------------------------------------------
		//	tbody
		//--------------------------------------------------------------

		var tbodyElement = document.createElement('tbody');

		var features = [];

		//--------------------------------------------------------------
		//	analysiere Linie fuer Linie
		//--------------------------------------------------------------

		var phi1    = this.math.toRadian (lonlats[0].lat);
		var lambda1 = this.math.toRadian (lonlats[0].lon);

		var sumOrtho = 0;
		var sumLoxo  = 0;

		for (var i=1; i<lonlats.length; i++) {

			var phi2    = this.math.toRadian (lonlats[i].lat);
			var lambda2 = this.math.toRadian (lonlats[i].lon);

			//------------------------------------------------------
			//	Sphaerische Berechnung
			//------------------------------------------------------

			var spheric = this.math.sphericalInverse({
				phi1: phi1, lambda1: lambda1,
				phi2: phi2, lambda2: lambda2
			});

			var sphericDistance = spheric.delta * this.EARTH_RADIUS;

			//------------------------------------------------------
			//	Loxodomische Berechnung
			//------------------------------------------------------

			var loxodromic = this.math.loxodromicInverse({
				phi1: phi1, lambda1: lambda1,
				phi2: phi2, lambda2: lambda2
			});

			var loxodromicDistance = loxodromic.delta * this.EARTH_RADIUS;

			//------------------------------------------------------
			//	Erzeuge Linie
			//------------------------------------------------------

			var nSegments = Math.min (this.MAX_SEGMENTS_PER_LINE,
				Math.ceil (sphericDistance / this.MAX_SEGMENT_LENGTH));
			var points = [];

			points.push (new OpenLayers.Geometry.Point (
				this.math.toDegrees(lambda1),
				this.math.toDegrees(phi1)));

			var lastLambda = lambda1;

			for (var s=1; s<=nSegments; s++) {

				var pos = this.math.sphericalForward ({
					phi1:    phi1,
					lambda1: lambda1,
					alpha1:  spheric.alpha1,
					delta:   spheric.delta * s / nSegments
				});

				if (lastLambda * pos.lambda2 < 0) { // wrap date line

					var geometry = new OpenLayers.Geometry.LineString(points);
					features.push (new OpenLayers.Feature.Vector(
						geometry, null, this.lineStyle));
					points = [];
				}
				lastLambda = pos.lambda2;

				points.push (new OpenLayers.Geometry.Point (
					this.math.toDegrees(pos.lambda2),
					this.math.toDegrees(pos.phi2)));
			}

			geometry = new OpenLayers.Geometry.LineString(points);
			features.push (new OpenLayers.Feature.Vector(geometry, null, this.lineStyle));

			// var ellipsoideDistance = geometry.getGeodesicLength();
			// if (isNaN(ellipsoideDistance)) window.alert ('No geodesic length: ' + points);

			//------------------------------------------------------
			//	Tabellenzeile
			//------------------------------------------------------

			tr = document.createElement('tr');
			this.appendTD(tr, this.kompassrichtung(spheric.alpha1));
			this.appendTD(tr, this.formatDirection(spheric.alpha1), 'right');
			//this.appendTD(tr, this.formatDistance(ellipsoideDistance), 'right');
			this.appendTD(tr, this.formatDistance(sphericDistance), 'right');
			this.appendTD(tr, this.kompassrichtung(spheric.alpha2+Math.PI));
			this.appendTD(tr, this.formatDirection(spheric.alpha2+Math.PI), 'right');
			this.appendTD(tr, this.kompassrichtung(loxodromic.alpha1));
			this.appendTD(tr, this.formatDirection(loxodromic.alpha1), 'right');
			this.appendTD(tr, this.formatDistance(loxodromicDistance), 'right');
			tbodyElement.appendChild(tr);

			sumLoxo  += loxodromicDistance;
			//sumOrtho += ellipsoideDistance;
			sumOrtho += sphericDistance;

			//------------------------------------------------------
			//	Tabellenzeile
			//------------------------------------------------------

			phi1    = phi2;
			lambda1 = lambda2;
		}

		//--------------------------------------------------------------
		//	Tabellenfuss
		//--------------------------------------------------------------

		if (lonlats.length >= 3) {

			var tfootElement = document.createElement('tfoot');

			tr = document.createElement('tr');
			this.appendTD(tr);
			this.appendTD(tr);
			this.appendTD(tr, this.formatDistance(sumOrtho), 'right');
			this.appendTD(tr);
			this.appendTD(tr);
			this.appendTD(tr);
			this.appendTD(tr);
			this.appendTD(tr, this.formatDistance(sumLoxo), 'right');
			tfootElement.appendChild(tr);

			tableElement.appendChild (tfootElement);
		}

		tableElement.appendChild (tbodyElement);

		onSuccess (tableElement, features);
	},

	appendTD: function (tr, text, align) {

		var td = document.createElement('td');
		if (align) td.style.textAlign = align;
		if (text) td.appendChild(document.createTextNode(text));
		tr.appendChild(td);
	},

	lineStyle: {
		strokeColor: 'red',
		strokeWidth: 5,
		strokeOpacity: 0.7,
		graphicZIndex: 3
	},

	//-------------------------------------------------------------
	//	formatting functions
	//-------------------------------------------------------------

	formatDistance: function (value) {

		return value>=1000? (value/1000).toFixed(2) + ' km' : value.toFixed(0) + ' m';
	},

	formatDirection: function (phi) {

		var degrees = phi / Math.PI * 180;
		return Math.round(degrees + 720) %360 + '\xb0';
	},

	compassdirection: function (phi) {

		return ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
			'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW'
		][Math.round (phi * this.RHO / 22.5 + 32) % 16];
	},

	kompassrichtung: function (alpha) {
		return ['N', 'NNO', 'NO', 'ONO', 'O', 'OSO', 'SO', 'SSO',
			'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW'
		][Math.round (alpha/Math.PI * 8 + 32) % 16];
	},

	CLASS_NAME: 'GeodeticAnalyzer'
};

//------------------------------------------------------------------------------
//	$Id: geodetic.js,v 1.11 2022/01/28 20:16:13 wolf Exp $
//------------------------------------------------------------------------------
